import argparse
from pathlib import Path
import warnings

import sys

vespa_location = Path(__file__).resolve().parent.parent.parent
sys.path.append(str(vespa_location))

from vespa.scripts.embedding import run_embedding
from vespa.scripts.conspred import run_conspred
from vespa.scripts.logodds import run_logodds
from vespa.scripts.vespa_run import run_vespa
from vespa.predict.config import (CACHE_DIR, MODEL_PATH_DICT)


def setup_argparse():
    parser = argparse.ArgumentParser(
        description="A meta script to execute the components of VESPA sequentially to obtian predictions. "
                    "Based on the publication 'Embeddings from protein language models predict conservation and variant effects' (https://doi.org/10.21203/rs.3.rs-584804/v1) "
    )

    parser.add_argument(
        "fasta_input",
        type=Path,
        help="Fasta sequence file; Needs to be present for mutation generation ",
    )

    parser.add_argument(
        "--vespa",
        help="Run the VESPA model",
        default=False,
        action=argparse.BooleanOptionalAction,
    )
    parser.add_argument(
        "--vespal",
        help="Run the VESPAl model",
        default=True,
        action=argparse.BooleanOptionalAction,
    )

    # Optional argument
    parser.add_argument(
        "--vespa_output_directory",
        required=False,
        type=Path,
        help="Path to directory containing outputs produced by VESPA."
    )

    parser.add_argument(
        "--use_existing_embeddings",
        required=False,
        type=Path,
        help="Path to computed embeddings to use in h5 format. "
             "Does not contain the original sequence data, only the sequence id!",
    )

    # shared arguments
    # between logodds, vespa
    parser.add_argument(
        "-m",
        "--mutation_file",
        required=False,
        type=Path,
        help="Path to a txt file listing protein ids and their SAVs that should be checked. The file should contain one ProteinID_SAV per line. If not specified all mutations will be considered.",
    )
    parser.add_argument(
        "--one_based_mutations",
        required=False,
        action="store_true",
        help="Flag to determine if mutation file is one based. NOTE: all SAV strings generated bt this tool are one based.",
    )

    # shared arguments
    # between logodds, embedding
    parser.add_argument(
        "--prott5_weights_cache",
        required=False,
        type=Path,
        help="Path to directory for storing the weights of the language model ProtT5.",
    )
    return parser, {'vespa_output_directory': Path("./vespa_run_directory/"),
                    'prott5_weights_cache': CACHE_DIR}


def embedding_argparse(group):
    group.add_argument(
        "--embedding_output",
        required=False,
        type=Path,
        help="Output path to store the computed embeddings in h5 format. "
             "Does not contain the original sequence data, only the sequence id!",
    )

    # return default values for parser
    return {'embedding_output': "T5_embedding.h5"}


def logodds_argparse(group):
    group.add_argument(
        "--logodds_output",
        required=False,
        type=Path,
        help="Output path to store the computed logodds in h5 format. Does not contain the original sequence data (FromAA in SAV format not reconstructable without sequence)! Contains a -1 in the matrix if a specific mutation was not computed.",
    )

    group.add_argument(
        "-r",
        "--reconstruction_probas",
        default=False,
        action="store_true",
        help="Output the reconstrcution probabilites of T5 for a masked language model problem. The output is a 20-dimensional vector for all considered residue positions (all positions with at least one mutation or all positions if no file provided) giving the raw probability for each amino acid according to the MUTANT_ORDER.",
    )

    group.add_argument(
        "--reconstruction_output",
        required=False,
        type=Path,
        help="Output path to store the computed reconstruction probabilities in h5 format. Does not contain the original sequence data (FromAA in SAV format not reconstructable without sequence)!",
    )

    group.add_argument(
        "--single_csv",
        required=False,
        type=Path,
        help="Output path to store the computed logodds in a single *.csv file. The information is scored as SeqID_SAV;SAVscore.",
    )

    group.add_argument(
        "--csv_dir",
        required=False,
        type=Path,
        help="Output path to store the computed logodds in multiple csv files (one per sequence). The information is scored as SAV;SAVscore in files named as seqid.csv",
    )

    return {'logodds_output': "T5_logodds.h5",
            'reconstruction_output': "T5_reconstruction_probas.h5"}


def conspred_argparse(group):
    group.add_argument(
        "--conspred_output_prefix",
        required=False,
        type=Path,
        help="A prefix for output files. The output will be written at <prefix>_class.fasta and to ,<prefix>_probs.h5 depending on the selected prediction output.",
    )

    group.add_argument(
        "-c",
        "--checkpoint",
        required=False,
        type=Path,
        default=None,
        help="A path for the pre-trained checkpoint for the conservation CNN",
    )

    group.add_argument(
        "--output_probs",
        type=bool,
        default=True,
        action=argparse.BooleanOptionalAction,
        help="Output probabilities for all classes, not only class with highest probability. The probabilities are stored in an h5 file with a dataset per-protein of shape Lx20 (L being the protein length). This output is written to <output_prefix>_probs.h5)",
    )

    group.add_argument(
        "--output_classes",
        type=bool,
        default=False,
        action=argparse.BooleanOptionalAction,
        help="Output the conservation class prediction per residue in a fasta-like format with comma-separated per-residue classes. The output is written to <output_prefix>_class.fast)",
    )

    return {'conspred_output_prefix': "conspred"}


def vespa_argparse(group):
    group.add_argument(
        "--vespa_pred_output",
        required=False,
        type=Path,
        nargs="?",
        help="Path to output csv file."
    )

    group.add_argument(
        "--no_csv",
        help="Disable csv output",
        required=False,
        action="store_true",
    )

    group.add_argument(
        "--h5_output",
        type=Path,
        help="Write output to a h5 file that contains Lx20 matrices per sequence per model; -1 means no score computed",
        default=None,
    )

    return {'vespa_pred_output': "output/"}


def main():
    parser, default_args = setup_argparse()
    embedding_group = parser.add_argument_group(title='embedding_args')
    logodds_group = parser.add_argument_group(title='logodss_args')
    conspred_group = parser.add_argument_group(title='conspred_args')
    vespa_group = parser.add_argument_group(title='vespa_args')

    default_emb_args = embedding_argparse(embedding_group)
    default_logodds_args = logodds_argparse(logodds_group)
    default_conspred_args = conspred_argparse(conspred_group)
    default_vespa_args = vespa_argparse(vespa_group)

    args = parser.parse_args()

    vespa_output_directory = args.vespa_output_directory.with_suffix('') if args.vespa_output_directory \
        else default_args['vespa_output_directory']
    vespa_output_directory.mkdir(parents=True, exist_ok=True)
    print(f'Save vespa output to {vespa_output_directory.resolve()}')

    if not args.prott5_weights_cache:
        warnings.warn("Memory Warning! Downloading the weights of ProtT5 due to the absence of a previous download!")
    cache_dir = args.prott5_weights_cache or vespa_output_directory.joinpath(default_args['prott5_weights_cache'])

    embedding_arguments = {
        'seq_path': args.fasta_input,
        'out_path':
            args.embedding_output.with_suffix('.h5') if args.embedding_output
            else vespa_output_directory.joinpath(default_emb_args['embedding_output']),
        'cache_dir': cache_dir
    }

    conspred_arguments = {
        'seq_path':
            args.use_existing_embeddings.with_suffix('.h5') if args.use_existing_embeddings
            else embedding_arguments['out_path'],
        'checkpoint_path':
            args.checkpoint or Path(vespa_location.joinpath(MODEL_PATH_DICT["CONSCNN"])),
        'out_prefix':
            args.conspred_output_prefix.with_suffix('') if args.conspred_output_prefix
            else vespa_output_directory.joinpath(default_conspred_args['conspred_output_prefix']),
        'write_probs': args.output_probs,
        'write_classes': args.output_classes
    }

    logodds_arguments = {
        'seq_path': args.fasta_input,
        'out_path':
            args.logodds_output.with_suffix('.h5') if args.logodds_output
            else vespa_output_directory.joinpath(default_logodds_args['logodds_output']),
        'mutation_file': args.mutation_file,
        'one_based_mutations': args.one_based_mutations,
        'reconstruction_probas': args.reconstruction_probas,
        'reconstruction_output':
            args.reconstruction_output.with_suffix('.h5') if args.reconstruction_output
            else vespa_output_directory.joinpath(default_logodds_args['reconstruction_output']),
        'single_csv': args.single_csv,
        'csv_dir': args.csv_dir,
        'cache_dir': cache_dir
    }

    vespa_arguments = {
        'seq_path': args.fasta_input,
        'out_path':
            args.vespa_pred_output.with_suffix('') if args.vespa_pred_output
            else vespa_output_directory.joinpath(default_vespa_args['vespa_pred_output']),
        'vespa': args.vespa,
        'vespal': args.vespal,
        'cons_input': Path(str(conspred_arguments['out_prefix']) + "_probs.h5"),
        'mutation_file': args.mutation_file,
        'one_based_mutations': args.one_based_mutations,
        'no_csv': args.no_csv,
        'h5_output': args.h5_output,
        'T5_input': logodds_arguments['out_path'],
    }

    if not vespa_arguments['vespa'] and not vespa_arguments['vespal']:
        raise RuntimeError(
            "Requested to run no vespa-model. Please at least allow one model! E.g. `--vespa` or `--vespal`"
        )

    if not args.use_existing_embeddings:
        warnings.warn('Generating embeddings might not be possible due to a GPU without enough memory!')
        run_embedding(**embedding_arguments)
    run_conspred(**conspred_arguments)
    if vespa_arguments['vespa']:
        run_logodds(**logodds_arguments)
    run_vespa(**vespa_arguments)


if __name__ == '__main__':
    main()



__author__ = "Tobias O, Michael H., Celine M., Duc Anh L."
__copyright__ = "Copyright 2022, Rostlab"
__license__ = "AGPL-3.0-or-later"
__version__ = "1.0.0"
__maintainer__ = ["Duc Anh L.", "Tobias O."]
__email__ = ""
__status__ = "Production"



#!/usr/bin/env python
"""
 Copyright (C) 2021 Rostlab
 
 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU Affero General Public License as
 published by the Free Software Foundation, either version 3 of the
 License, or (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Affero General Public License for more details.
 
 You should have received a copy of the GNU Affero General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

# -*- coding: utf-8 -*-

"""
File: conspred
{Description}
"""

# Default Imports
from pathlib import Path
import argparse
import h5py


# Lib Imports
import torch.utils.data
from vespa.predict.config import MODEL_PATH_DICT, VERBOSE

# Module Imports
from vespa.predict.conspred import ProtT5Cons, get_dataloader


def create_arg_parser():
    """ "Creates and returns the ArgumentParser object."""

    # Instantiate the parser
    parser = argparse.ArgumentParser(
        description=(
            "consurf_infer.py predicts conservation in [0,8] for a given text "
            + " file containing sequence(s) in FASTA-format."
        )
    )

    # Required positional argument
    parser.add_argument(
        "input",
        type=Path,
        help="A path to a h5 embedding file, containing per-residue ProtT5 embeddings.",
    )

    # Optional positional argument
    parser.add_argument(
        "-o",
        "--output_prefix",
        required=False,
        default="./conspred",
        type=str,
        help="A prefix for output files. The output will be written at <prefix>_class.fasta and to ,<prefix>_probs.h5 depending on the selected prediction output.",
    )

    # Optional positional argument
    parser.add_argument(
        "-c",
        "--checkpoint",
        required=False,
        type=Path,
        default=None,
        help="A path for the pre-trained checkpoint for the conservation CNN",
    )

    # Optional argument
    parser.add_argument(
        "--output_probs",
        type=bool,
        default=True,
        action=argparse.BooleanOptionalAction,
        help="Output probabilities for all classes, not only class with highest probability. The probabilities are stored in an h5 file with a dataset per-protein of shape Lx20 (L being the protein length). This output is written to <output_prefix>_probs.h5)",
    )

    # Optional argument
    parser.add_argument(
        "--output_classes",
        type=bool,
        default=False,
        action=argparse.BooleanOptionalAction,
        help="Output the conservation class prediction per residue in a fasta-like format with comma-separated per-residue classes. The output is written to <output_prefix>_class.fast)",
    )

    return parser


def run_conspred(seq_path, checkpoint_path, write_probs, write_classes, out_prefix):
    try:
        if VERBOSE:
            print(f" Start Conservation Prediction ".center(80, "#"))
        embeddings = h5py.File(str(seq_path.resolve()), "r")
        data_loader = get_dataloader(embeddings, batch_size=128)
        if VERBOSE:
            print(f" Load model! ")
        conspred = ProtT5Cons(checkpoint_path)
        if VERBOSE:
            print(f" Predict Conservation! ")
        predictions = conspred.conservation_prediction(
            data_loader, prob_return=write_probs, class_return=write_classes
        )

        out_class = Path(str(out_prefix) + "_class.fasta")
        out_probs = Path(str(out_prefix) + "_probs.h5")
        out_class.parent.mkdir(parents=True, exist_ok=True)

        if write_classes:
            conspred.write_cons_class_pred(predictions, out_class)
        if write_probs:
            conspred.write_probabilities(predictions, out_probs)

        if VERBOSE:
            print(f">> Finished Conservation Prediction!")
    finally:
        embeddings.close()


def main():
    parser = create_arg_parser()
    args = parser.parse_args()

    arguments = {
        "seq_path": args.input,
        "checkpoint_path": args.checkpoint
        if args.checkpoint
        else Path(MODEL_PATH_DICT["CONSCNN"]),
        "out_prefix": args.output_prefix,
        "write_probs": args.output_probs,
        "write_classes": args.output_classes,
    }

    run_conspred(**arguments)


if __name__ == "__main__":
    main()



import argparse
from vespa.predict.embedding import *
from vespa.predict.config import VERBOSE, CACHE_DIR


def setup_argparse():
    parser = argparse.ArgumentParser(
        description="embedding.py creates T5 embeddings for a given text "
                    + " file containing sequence(s) in FASTA-format."
    )

    # Required argument
    parser.add_argument(
        "input",
        type=Path,
        metavar="fasta-file",
        help="A path to a fasta-formatted text file containing protein sequence(s).",
    )

    # Optional argument
    parser.add_argument(
        "-o",
        "--output",
        required=False,
        type=Path,
        default=Path("T5_embedding.h5"),
        help="Output path to store the computed embeddings in h5 format. "
             "Does not contain the original sequence data, only the sequence id!",
    )

    parser.add_argument(
        "--prott5_weights_cache",
        required=False,
        type=Path,
        default=Path(CACHE_DIR),
        help="Path to directory for storing the weights of the language model ProtT5.",
    )

    return parser


def run_embedding(seq_path, out_path, cache_dir):
    if VERBOSE:
        print(f' Start Embedding of {seq_path.stem} '.center(80, '#'))
    t5_emb = T5_Embed(cache_dir=cache_dir)
    t5_emb.embed_from_fasta(fasta_path=seq_path, output_path=out_path)
    if VERBOSE:
        print(f'>> Finished Embeddings!')


def main():
    parser = setup_argparse()
    args = parser.parse_args()

    arguments = {
        'seq_path': args.input,
        'out_path': args.output,
        'cache_dir': args.prott5_weights_cache
    }
    run_embedding(**arguments)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
 Copyright (C) 2021 Rostlab
 
 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU Affero General Public License as
 published by the Free Software Foundation, either version 3 of the
 License, or (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Affero General Public License for more details.
 
 You should have received a copy of the GNU Affero General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

# -*- coding: utf-8 -*-

"""
File: conspred
Conservation Prediction for embeddings
"""

# Default Imports
from collections import namedtuple
import enum
from pathlib import Path

# Lib Imports
import torch
import torch.utils.data
import torch.nn as nn
from torch.nn.utils.rnn import pad_sequence
import h5py
from tqdm import tqdm

# Module Imports
from vespa.predict.config import CACHE_DIR, DEVICE, VERBOSE


ConservationOut = namedtuple("ConservationOut", ["probability_out", "class_out"])

# Convolutional neural network (two convolutional layers)
class ConsCNN(nn.Module):
    def __init__(self):
        super(ConsCNN, self).__init__()
        n_features = 1024
        bottleneck_dim = 32
        n_classes = 9
        dropout_rate = 0.25
        self.classifier = nn.Sequential(
            nn.Conv2d(
                n_features, bottleneck_dim, kernel_size=(7, 1), padding=(3, 0)
            ),  # 7x32
            nn.ReLU(),
            nn.Dropout(dropout_rate),
            nn.Conv2d(bottleneck_dim, n_classes, kernel_size=(7, 1), padding=(3, 0)),
        )
        self.softmax = nn.Softmax(dim=1)

    def forward(self, x):
        """
        L = protein length
        B = batch-size
        F = number of features (1024 for embeddings)
        N = number of classes (9 for conservation)
        """
        # IN: X = (B x L x F); OUT: (B x F x L, 1)
        x = x.permute(0, 2, 1).unsqueeze(dim=-1)
        Yhat_consurf = self.classifier(x)  # OUT: Yhat_consurf = (B x N x L x 1)
        # IN: (B x N x L x 1); OUT: ( B x L x N )
        Yhat_consurf = Yhat_consurf.squeeze(dim=-1)
        return Yhat_consurf

    def extract_probabilities(self, Yhat):
        return self.softmax(Yhat)

    def extract_conservation_score(self, Yhat):
        return torch.max(Yhat.data, dim=1)[1]  # get index of most reliable class prediction


class BatchCollator(object):
    def __call__(self, batch, ignore_idx=-100):
        # batch is a list of the samples returned by your __get_item__ method in your CustomDataset
        pdb_ids, X, seq_lens = zip(*batch)
        X = pad_sequence(X, batch_first=True, padding_value=ignore_idx)
        return (list(pdb_ids), X, seq_lens)


class SequenceDataset(torch.utils.data.Dataset):
    def __init__(self, h5file):
        self.interface = h5file
        self.keys = [key for key in self.interface]
        self.data_len = len(self.keys)  # number of samples in the set

    def __len__(self):
        return self.data_len

    def __getitem__(self, index):
        pdb_id = self.keys[index]
        x = torch.from_numpy(self.interface[pdb_id][:]).to(torch.float32)
        return (pdb_id, x, x.shape[0])



def get_dataloader(customdata, batch_size):
    # Create dataloaders with collate function
    collator = BatchCollator()
    dataset = SequenceDataset(customdata)
    return torch.utils.data.DataLoader(
        dataset=dataset,
        batch_size=batch_size,
        shuffle=False,
        drop_last=False,
        num_workers=3,
        collate_fn=collator,
    )
class ProtT5Cons:
    def __init__(self, checkpoint_path):
        self.model = ConsCNN().to(DEVICE)
        self.predictor = self.load_checkpoint(checkpoint_path)

    def load_checkpoint(self, checkpoint_path: Path):
        state = torch.load(checkpoint_path, map_location=DEVICE)
        self.model.load_state_dict(state["state_dict"])
        self.model = self.model.eval()
        return self.model

    def conservation_prediction(
        self, data_loader, prob_return=True, class_return=True
    ) -> ConservationOut:
        if prob_return == class_return == False:
            raise RuntimeError(
                "Method needs to either return probabilities or conservation class or both."
            )

        probability_out = dict() if prob_return else None
        class_out = dict() if class_return else None
        ret = ConservationOut(probability_out=probability_out, class_out=class_out)

        for i, (pdb_ids, X, lens) in enumerate(data_loader):
            X = X.to(DEVICE)
            with torch.no_grad():
                Yhat = self.predictor(X)
                if prob_return:
                    prob = self.model.extract_probabilities(Yhat).detach().cpu().numpy()
                if class_return:
                    cls = (
                        self.model.extract_conservation_score(Yhat)
                        .detach()
                        .cpu()
                        .numpy()
                    )
            for idx, pdb_id in enumerate(pdb_ids):
                if prob_return:
                    ret.probability_out[pdb_id] = prob[idx, :, : lens[idx]]
                if class_return:
                    ret.class_out[pdb_id] = cls[idx, : lens[idx]]
        return ret

    @staticmethod
    def write_cons_class_pred(predictions: ConservationOut, out_path):
        if predictions.class_out is None:
            raise RuntimeError("Class output not present!")
        with open(out_path, "w+") as out_f:
            for seq_id, cons_cls in predictions.class_out.items():
                out_f.write(f">{seq_id}\n{','.join([str(j) for j in cons_cls])}\n")

    @staticmethod
    def write_probabilities(predictions: ConservationOut, out_path):
        if predictions.probability_out is None:
            raise RuntimeError("Probability output not present!")
        with h5py.File(str(out_path), "w") as hf:
            for sequence_id, embedding in predictions.probability_out.items():
                # noinspection PyUnboundLocalVariable
                hf.create_dataset(sequence_id, data=embedding)
        return None



#!/usr/bin/env python
"""
 Copyright (C) 2021 Rostlab
 
 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU Affero General Public License as
 published by the Free Software Foundation, either version 3 of the
 License, or (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Affero General Public License for more details.
 
 You should have received a copy of the GNU Affero General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

# -*- coding: utf-8 -*-

from pathlib import Path
import torch

VESPA_LOCATION = Path(__file__).resolve().parent.parent.parent

VESPA = "VESPA"
VESPAL = "VESPAl"

MODEL_PATH_DICT = {
    VESPA: Path(VESPA_LOCATION.joinpath("models/VESPA-10LR_Cons_Blsm_Prob.pkl")),
    VESPAL: Path(VESPA_LOCATION.joinpath("models/VESPAl-10LR_Cons_Blsm.pkl")),
    "CONSCNN": Path(VESPA_LOCATION.joinpath("models/ProtT5cons_checkpoint.pt")),
}

OUTPUT_MAP_NAME = "map.json"

# Special character preprended to all "start of span" (for us: every AA as we only do 1-token spans)
# https://huggingface.co/transformers/v3.1.0/_modules/transformers/tokenization_t5.html
SPIECE_UNDERLINE = "▁"

CACHE_DIR = "./cache"

TRANSFORMER_LINK = "Rostlab/prot_t5_xl_uniref50"

REPLACE_RARE_AA = True

MUTANT_ORDER = "ALGVSREDTIPKFQNYMHWC" if REPLACE_RARE_AA else "ALGVSREDTIPKFQNYMHWCUZO"
REVERSE_MUTANT_ORDER = {item: key for key, item in enumerate(MUTANT_ORDER)}

CSV_SEP = ";"
RESULT_PREC = -1  # -1 for all comma positions, positive int for cut precision

NUM_MUTATION_SCORES = 9

VERBOSE = True

DEVICE = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
EMBEDDING_HALF_PREC = True

EMBED, LOGODDS = 0, 1
EMB_MAX_SEQ_LEN, EMB_MAX_RESIDUES, EMB_MAX_BATCH, EMB_STORE_FREQ = 600, 8000, 5, 200


"""
 Copyright (C) 2021 Rostlab
 
 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU Affero General Public License as
 published by the Free Software Foundation, either version 3 of the
 License, or (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Affero General Public License for more details.
 
 You should have received a copy of the GNU Affero General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import argparse

from vespa.predict.config import VERBOSE, CACHE_DIR
from vespa.predict.logodds import *
from vespa.predict import utils


def create_arg_parser():
    """ "Creates and returns the ArgumentParser object."""

    # Instantiate the parser
    parser = argparse.ArgumentParser(
        description=(
            "logodds.py creates T5 logodds for a given text "
            + " file containing sequence(s) in FASTA-format."
        )
    )

    # Required argument
    parser.add_argument(
        "input",
        type=Path,
        metavar="fasta-file",
        help="A path to a fasta-formatted text file containing protein sequence(s).",
    )

    # Optional argument
    parser.add_argument(
        "-o",
        "--output",
        required=False,
        type=Path,
        default=Path("T5_logodds.h5"),
        help="Output path to store the computed logodds in h5 format. Does not contain the original sequence data (FromAA in SAV format not reconstructable without sequence)! Contains a -1 in the matrix if a specific mutation was not computed.",
    )

    parser.add_argument(
        "-r",
        "--reconstruction_probas",
        default=False,
        action="store_true",
        help="Output the reconstrcution probabilites of T5 for a masked language model problem. The output is a 20-dimensional vector for all considered residue positions (all positions with at least one mutation or all positions if no file provided) giving the raw probability for each amino acid according to the MUTANT_ORDER.",
    )

    parser.add_argument(
        "--reconstruction_output",
        type=Path,
        default=Path("T5_reconstruction_probas.h5"),
        help="Output path to store the computed reconstruction probabilities in h5 format. Does not contain the original sequence data (FromAA in SAV format not reconstructable without sequence)!",
    )

    parser.add_argument(
        "--single_csv",
        required=False,
        type=Path,
        help="Output path to store the computed logodds in a single *.csv file. The information is scored as SeqID_SAV;SAVscore.",
    )

    parser.add_argument(
        "--csv_dir",
        required=False,
        type=Path,
        help="Output path to store the computed logodds in multiple csv files (one per sequence). The information is scored as SAV;SAVscore in files named as seqid.csv",
    )

    parser.add_argument(
        "-m",
        "--mutation_file",
        required=False,
        type=Path,
        help="Path to a txt file listing protein ids and their SAVs that should be checked. The file should contain one ProteinID_SAV per line. If not specified all mutations will be considered.",
    )
    parser.add_argument(
        "--one_based_mutations",
        required=False,
        action="store_true",
        help="Flag to determine if mutation file is one based. NOTE: all SAV strings generated bt this tool are one based.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        required=False,
        action="store_true",
        help="Control the output of the script.",
    )

    parser.add_argument(
        "--prott5_weights_cache",
        required=False,
        type=Path,
        default=Path(CACHE_DIR),
        help="Path to directory for storing the weights of the language model ProtT5.",
    )

    return parser


def run_logodds(seq_path, out_path, mutation_file, one_based_mutations, reconstruction_probas,
                reconstruction_output, single_csv, csv_dir, cache_dir):
    if VERBOSE:
        print(f' Start Logodds Computation '.center(80, '#'))
    seq_dict = utils.parse_fasta_input(seq_path)
    mutation_gen = utils.MutationGenerator(
        sequence_dict=seq_dict,
        file_path=mutation_file,
        one_based_file=one_based_mutations,
    )

    dmiss_data, probas = create_log_odds(
        seq_dict, mutation_gen, cache_dir, reconstruction_probas
    )

    # TODO maybe save logodds more frequently?
    T5_condProbas.write_single_h5(dmiss_data, out_path)

    if reconstruction_probas:
        T5_condProbas.write_single_h5(probas, reconstruction_output)

    if single_csv:
        T5_condProbas.write_single_csv(dmiss_data, single_csv, mutation_gen)
    if csv_dir:
        T5_condProbas.write_csv_dir(dmiss_data, csv_dir, mutation_gen)

    if VERBOSE:
        print(f'>> Finished Logodds Computation!')


def create_log_odds(seq_dict, mutation_gen, cache_dir, extract_probas=False):
    t5_condProbas = T5_condProbas(cache_dir=cache_dir)
    if VERBOSE:
        print("Running DMISS (Deep mutational in-silico scanning.)")
    proba_dict = t5_condProbas.get_proba_dict(seq_dict, mutation_gen)
    dmiss_data = t5_condProbas.get_log_odds(proba_dict)
    probas = None
    if extract_probas:
        probas = t5_condProbas.get_rec_probabilities(proba_dict)

    return dmiss_data, probas


def main():
    parser = create_arg_parser()
    args = parser.parse_args()

    VERBOSE = args.verbose

    arguments = {
        'seq_path': args.input,
        'out_path': args.output,
        'mutation_file': args.mutation_file,
        'one_based_mutations': args.one_based_mutations,
        'reconstruction_probas': args.reconstruction_probas,
        'reconstruction_output': args.reconstruction_output,
        'single_csv': args.single_csv,
        'csv_dir': args.csv_dir,
        'cache_dir': args.prott5_weights_cache
    }
    run_logodds(**arguments)


if __name__ == '__main__':
    main()



import numpy
import torch
import h5py
from tqdm import tqdm
from pathlib import Path

from vespa.predict.config import (
    DEVICE, CACHE_DIR, VERBOSE,
    EMBED, EMB_MAX_SEQ_LEN, EMB_MAX_RESIDUES, EMB_MAX_BATCH, EMB_STORE_FREQ
)
from vespa.predict.utils import parse_fasta_input
from vespa.predict.utils_t5 import ProtT5


class T5_Embed:
    def __init__(self, cache_dir):
        self.prott5 = ProtT5(cache_dir)
        self.saving_pattern = 'w'

    def embed_from_fasta(self, fasta_path, output_path):
        self.saving_pattern = 'w'
        if VERBOSE:
            print('Load model: ProtT5')
        self.model, self.tokenizer = self.prott5.get_model(EMBED)
        if VERBOSE:
            print('Compute embeddings!')
        self.get_embeddings(fasta_path, output_path)

    def embedding_init(self, fasta_path):
        seq_dict = parse_fasta_input(fasta_path)
        seq_dict = sorted(seq_dict.items(), key=lambda kv: len(seq_dict[kv[0]]), reverse=True)
        return seq_dict

    def process_batch(self, batch, emb_dict):
        pdb_ids, seqs, seq_lens = zip(*batch)

        token_encoding = self.tokenizer(seqs, add_special_tokens=True, padding='longest', return_tensors="pt")
        input_ids = token_encoding['input_ids'].to(DEVICE)
        attention_mask = token_encoding['attention_mask'].to(DEVICE)

        try:
            # batch-size x seq_len x embedding_dim
            with torch.no_grad():
                embedding_repr = self.model(input_ids, attention_mask=attention_mask)
        except RuntimeError:
            print("RuntimeError for {} (L={})".format(pdb_ids, seq_lens))
            return emb_dict

        new_emb_dict = dict()
        for batch_idx, identifier in enumerate(pdb_ids):
            s_len = seq_lens[batch_idx]
            emb = embedding_repr.last_hidden_state[batch_idx, :s_len]
            new_emb_dict[identifier] = emb.detach().cpu().numpy().squeeze()

        if new_emb_dict:
            emb_dict.update(new_emb_dict)
        return emb_dict

    def save_embeddings(self, output_path, emb_dict):
        Path(str(output_path.absolute())).parent.mkdir(parents=True, exist_ok=True)
        with h5py.File(str(output_path.absolute()), self.saving_pattern) as hf:
            for sequence_id, embedding in emb_dict.items():
                hf.create_dataset(sequence_id, data=embedding)
        self.saving_pattern = 'a'

    def get_embeddings(self, fasta_path, output_path):
        seq_dict = self.embedding_init(fasta_path)

        emb_dict = dict()
        batch, n_res_batch = [], 0

        for seq_idx, (pdb_id, seq) in tqdm(enumerate(seq_dict, 1), total=len(seq_dict)):
            seq_len = len(seq)
            seq = ' '.join(list(seq))

            if seq_len >= EMB_MAX_SEQ_LEN:
                emb_dict = self.process_batch([(pdb_id, seq, seq_len)], emb_dict)
            else:
                if len(batch) >= EMB_MAX_BATCH or n_res_batch >= EMB_MAX_RESIDUES:
                    emb_dict = self.process_batch(batch, emb_dict)
                    batch = []
                    n_res_batch = 0

                batch.append((pdb_id, seq, seq_len))
                n_res_batch += seq_len

            if len(emb_dict) > EMB_STORE_FREQ:
                self.save_embeddings(output_path, emb_dict)
                emb_dict = dict()

        if batch:
            emb_dict = self.process_batch(batch, emb_dict)

        if emb_dict:
            self.save_embeddings(output_path, emb_dict)



#!/usr/bin/env python
"""
 Copyright (C) 2021 Rostlab
 
 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU Affero General Public License as
 published by the Free Software Foundation, either version 3 of the
 License, or (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Affero General Public License for more details.
 
 You should have received a copy of the GNU Affero General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

# -*- coding: utf-8 -*-

import argparse
import time
from pathlib import Path
from dataclasses import dataclass
from typing import Dict, List, Tuple

import numpy as np
import torch
import h5py
from tqdm import tqdm

from vespa.predict.config import (
    CACHE_DIR,
    CSV_SEP,
    MUTANT_ORDER,
    REPLACE_RARE_AA,
    REVERSE_MUTANT_ORDER,
    TRANSFORMER_LINK,
    SPIECE_UNDERLINE,
    VERBOSE,
    DEVICE,
    LOGODDS
)
from vespa.predict import utils
from vespa.predict.utils_t5 import ProtT5


if VERBOSE:
    print("Using device: {}".format(DEVICE))


@dataclass
class _ProbaVector:
    pos: int
    vector: np.ndarray
    mask: np.ndarray
    wt_vector_pos: int


class T5_condProbas:
    def __init__(self, cache_dir):
        self.prott5 = ProtT5(cache_dir)
        self.tokenizer = self.prott5.get_tokenizer()
        self.AAs = MUTANT_ORDER + "X"
        self.AA2class = {AA: idx for idx, AA in enumerate(self.AAs)}
        self.class2AA = {idx: AA for idx, AA in enumerate(self.AAs)}
        self.aaIdcs = [
            self.tokenizer.get_vocab()["{}{}".format(SPIECE_UNDERLINE, residue)]
            for residue in self.AAs
        ]
        self.softmax = torch.nn.Softmax(dim=0)

    def reconstruct_sequence(self, probs):
        return [self.class2AA[yhat] for yhat in probs.argmax(axis=1)]

    def measure_reconstruction_accuracy(self, wt_seq, probs):
        predictions = self.reconstruct_sequence(probs)
        correct = sum(
            [1 for idx, pred in enumerate(predictions) if pred == wt_seq[idx]]
        )
        return correct / len(wt_seq)

    def get_mask_prob(self, masked_seq, mut_idx):
        input_ids = self.tokenizer(masked_seq, return_tensors="pt").input_ids.to(DEVICE)
        try:  # returns a tuple with loss and logits
            with torch.no_grad():  # only logits is useful for us
                output = self.model(input_ids, labels=input_ids).logits
        except RuntimeError as e:
            print("RuntimeError for input with shape: {}".format(input_ids.shape))
            print(e)
            return None
        aa_logits = output[
            0, mut_idx, self.aaIdcs
        ]  # extract for the masked position only AA logits and potentially X if rare_AAs are replaced
        return self.softmax(aa_logits)

    def get_proba_dict(self, seq_dict, mutation_gen: utils.MutationGenerator):
        """
        Compute for all residues in a protein the conditional probabilities for reconstructing single, masked tokens.
        """

        self.model, self.tokenizer = self.prott5.get_model(LOGODDS)

        result_dict = dict()

        for identifier, position_list in tqdm(
            mutation_gen.generate_mutation_mask(include_x_pos=REPLACE_RARE_AA),
            desc="Extract Sequence Logodds",
            disable=not VERBOSE,
        ):

            wt_seq = seq_dict[identifier]
            seq_len = len(wt_seq)
            result_dict[identifier] = seq_len, list()

            for residue_idx, filter_mask in position_list.items():

                # Mask the residue in question
                masked_seq = list(wt_seq)
                masked_seq[
                    residue_idx
                ] = "<extra_id_0>"  # replace AA by special mask token used by T5
                masked_seq = " ".join(masked_seq)

                # Extract probabilities for all mutations
                aa_probas = self.get_mask_prob(masked_seq, residue_idx)
                if aa_probas is None:
                    print(
                        "Skipping {} (L={}) due to RunTimeError.".format(
                            identifier, seq_len
                        )
                    )
                    continue
                result_vec = _ProbaVector(
                    pos=residue_idx,
                    vector=aa_probas.detach().cpu().numpy().squeeze(),
                    mask=filter_mask,
                    wt_vector_pos=self.AA2class[wt_seq[residue_idx]],
                )
                result_dict[identifier][1].append(result_vec)

        return result_dict

    @staticmethod
    def get_rec_probabilities(
        probabiliy_dict: Dict[str, Tuple[int, List[_ProbaVector]]]
    ):
        result_dict = dict()
        for id in probabiliy_dict:
            seq_len = probabiliy_dict[id][0]
            array = np.ones((seq_len, len(MUTANT_ORDER))) * -1
            for prob_vec in probabiliy_dict[id][1]:
                # Remove X reconstruction prob as that probability makes little sense
                vec = prob_vec.vector[:-1]
                array[prob_vec.pos] = vec
            result_dict[id] = array
        return result_dict

    @staticmethod
    def get_log_odds(probabiliy_dict: Dict[str, Tuple[int, List[_ProbaVector]]]):
        result_dict = dict()
        for id in probabiliy_dict:
            seq_len = probabiliy_dict[id][0]
            array = np.ones((seq_len, len(MUTANT_ORDER))) * -1
            for prob_vec in probabiliy_dict[id][1]:
                log_vector = np.log(prob_vec.vector)
                log_vector -= log_vector[prob_vec.wt_vector_pos]
                # Remove X reconstruction prob as that probability makes little sense
                vec = log_vector * prob_vec.mask
                vec = vec[:-1]
                array[prob_vec.pos] = vec
            result_dict[id] = array
        return result_dict

    @staticmethod
    def write_single_h5(dmiss_data, h5_path):
        with h5py.File(str(h5_path), "w") as hf:
            for seq_id, log_odds in dmiss_data.items():
                # noinspection PyUnboundLocalVariable
                hf.create_dataset(seq_id, data=log_odds)
        return None

    @staticmethod
    def write_single_csv(
        dmiss_data, csv_path: Path, mutation_gen: utils.MutationGenerator
    ):
        with csv_path.open("w") as out_file:
            out_file.write(f"SeqID_SAV{CSV_SEP}SAV_score\n")
            for id, mutations in T5_condProbas.__gen_mutation_lists(
                dmiss_data, mutation_gen
            ):
                lines = [
                    f"{id}_{sav}{CSV_SEP}{score}\n" for sav, score in mutations.items()
                ]
                out_file.writelines(lines)

    @staticmethod
    def __gen_mutation_lists(dmiss_data, mutation_gen: utils.MutationGenerator):
        AA2idx = mutation_gen._get_aa2idx()
        mutation_dict = dict()
        old_id = None
        for id, pos0idx, fromAA, toAA in mutation_gen.generate_all_mutations():
            if old_id is None:
                old_id = id
            if id != old_id:
                yield old_id, mutation_dict
                old_id = id
                mutation_dict = dict()

            mutation_handle = utils.generate_sav_str(fromAA, pos0idx, toAA)
            mutation_score = dmiss_data[id][pos0idx, AA2idx[toAA]]
            mutation_dict[mutation_handle] = mutation_score
        yield old_id, mutation_dict

    @staticmethod
    def write_csv_dir(dmiss_data, mutation_dir: Path, mutation_gen):
        if mutation_dir.exists() and not mutation_dir.is_dir():
            raise ValueError("Output path must be a directory!")
        else:
            mutation_dir.mkdir(parents=True, exist_ok=True)
        for id, mutations in T5_condProbas.__gen_mutation_lists(
            dmiss_data, mutation_gen
        ):
            csv_path = mutation_dir / f"{id}.csv"
            with csv_path.open("w") as out_file:
                out_file.write(f"SAV{CSV_SEP}SAV_score\n")
                lines = [f"{sav}{CSV_SEP}{score}\n" for sav, score in mutations.items()]
                out_file.writelines(lines)



"""
 Copyright (C) 2021 Rostlab
 
 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU Affero General Public License as
 published by the Free Software Foundation, either version 3 of the
 License, or (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Affero General Public License for more details.
 
 You should have received a copy of the GNU Affero General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""




"""
 Copyright (C) 2021 Rostlab
 
 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU Affero General Public License as
 published by the Free Software Foundation, either version 3 of the
 License, or (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Affero General Public License for more details.
 
 You should have received a copy of the GNU Affero General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import argparse

from vespa.predict.vespa import *
from vespa.predict.config import VERBOSE
from vespa.predict import utils


def setup_argparse():
    parser = argparse.ArgumentParser(
        description="A cli tool for VESPA predictions. Based on the publication 'Embeddings from protein language models predict conservation and variant effects' (https://doi.org/10.21203/rs.3.rs-584804/v1) "
    )
    parser.add_argument(
        "--vespa",
        help="Run the VESPA model",
        default=True,
        action=argparse.BooleanOptionalAction,
    )
    parser.add_argument(
        "--vespal",
        help="Run the VESPAl model",
        default=True,
        action=argparse.BooleanOptionalAction,
    )
    parser.add_argument(
        "cons_input",
        type=Path,
        help="Conservation prediction for the relevant sequences.",
    )
    parser.add_argument(
        "--T5_input",
        nargs="?",
        type=Path,
        help="T5 log odds for the relevant sequences. Needs to be present to run VESPA",
    )
    parser.add_argument(
        "fasta_file",
        type=Path,
        help="Fasta sequence file; Needs to be present for mutation generation ",
    )
    parser.add_argument(
        "--output",
        type=Path,
        nargs="?",
        help="Path to output csv file.",
        default=Path("./output/"),
    )

    parser.add_argument(
        "--no_csv",
        help="Disable csv output",
        required=False,
        action="store_true",
    )

    parser.add_argument(
        "--h5_output",
        type=Path,
        help="Write output to a h5 file that contains Lx20 matrices per sequence per model; -1 means no score computed",
        default=None,
    )

    parser.add_argument(
        "-m",
        "--mutation_file",
        required=False,
        type=Path,
        help="Path to a txt file listing protein ids and their SAVs that should be checked. The file should contain one ProteinID_SAV per line. If not specified all mutations will be considered.",
    )
    parser.add_argument(
        "--one_based_mutations",
        required=False,
        action="store_true",
        help="Flag to determine if mutation file is one based",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        required=False,
        action="store_true",
        help="Control the output of the script.",
    )

    return parser


def run_vespa(seq_path, out_path, vespa, vespal, cons_input, T5_input,
              mutation_file, one_based_mutations, no_csv, h5_output):
    if no_csv and not h5_output:
        raise RuntimeError(
            "Requested to run no vespa without output. Please specify --h5_output or remove --no_csv."
        )

    if not vespa and not vespal:
        raise RuntimeError(
            "Requested to run no vespa-model. Please at least allow one model! E.g. `--vespa` or `--vespal`"
        )

    if VERBOSE:
        print(f' Start Vespa Predictions '.center(80, '#'))
    seq_dict = utils.parse_fasta_input(seq_path)
    mutation_gen = utils.MutationGenerator(
        sequence_dict=seq_dict,
        file_path=mutation_file,
        one_based_file=one_based_mutations,
    )

    predictor = VespaPred(vespa=vespa, vespal=vespal)
    input = dict()
    input["conservations"] = predictor.parse_cons_input(cons_input)
    if vespa:
        input["logodds"] = predictor.parse_logodds_input(T5_input)
    output = predictor.generate_predictions(mutation_gen=mutation_gen, **input)

    if not no_csv:
        predictor.write_output_csv(output, out_path)
    if h5_output:
        seq2len = {key: len(seq_dict[key]) for key in seq_dict}
        predictor.write_output_h5(output, seq2len, h5_output)

    if VERBOSE:
        print(f'>> Finished Vespa Predictions!')


def main():
    parser = setup_argparse()
    args = parser.parse_args()

    arguments = {
        'seq_path': args.fasta_file,
        'out_path': args.output,
        'vespa': args.vespa,
        'vespal': args.vespal,
        'cons_input': args.cons_input,
        'mutation_file': args.mutation_file,
        'one_based_mutations': args.one_based_mutations,
        'no_csv': args.no_csv,
        'h5_output': args.h5_output,
        'T5_input': args.T5_input,
    }

    VERBOSE = args.verbose

    run_vespa(**arguments)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
 Copyright (C) 2021 Rostlab
 
 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU Affero General Public License as
 published by the Free Software Foundation, either version 3 of the
 License, or (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Affero General Public License for more details.
 
 You should have received a copy of the GNU Affero General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

# -*- coding: utf-8 -*-

from typing import Dict, List, Tuple
import pickle as pkl
from collections import defaultdict
from pathlib import Path
import json

from tqdm import tqdm
import h5py
import numpy as np
from Bio.Align import substitution_matrices

from vespa.predict.config import (
    MODEL_PATH_DICT,
    NUM_MUTATION_SCORES,
    OUTPUT_MAP_NAME,
    REVERSE_MUTANT_ORDER,
    VERBOSE,
    VESPAL,
    VESPA,
    CSV_SEP,
    MUTANT_ORDER,
)
from vespa.predict import utils


class VespaPred:
    def __init__(self, vespa=True, vespal=True) -> None:
        self.vespa_pred = vespa
        self.vespal_pred = vespal
        self.subst_matrix = self.__load_subst()
        self.models = self.__load_models()

    def __load_subst(self):
        return substitution_matrices.load("BLOSUM62")

    def write_output_csv(self, output, output_dir: Path):
        if VERBOSE:
            tqdm.write(f"Writing CSV Output to {output_dir.absolute()}")
        if output_dir.exists() and not output_dir.is_dir():
            raise ValueError("Output path must be a directory!")
        else:
            output_dir.mkdir(parents=True, exist_ok=True)
        header_str = f"Mutant{CSV_SEP}{CSV_SEP.join(list(self.models.keys()))}\n"
        map = dict()
        for idx, id in enumerate(output):
            out_path = output_dir / f"{idx}.csv"
            map[idx] = id
            with out_path.open("w") as out_file:
                out_file.write(header_str)
                for mutation in output[id]:
                    info, pred = mutation
                    out_file.write(
                        f"{info}{CSV_SEP}{CSV_SEP.join([utils.pred_to_str(pred[name]) for name in self.models])}\n"
                    )
        with (output_dir / OUTPUT_MAP_NAME).open("w") as map_file:
            json.dump(map, map_file, indent=4)

    def write_output_h5(self, output, seq_2_len: Dict[str, int], output_h5: Path):
        if VERBOSE:
            tqdm.write(f"Writing H5 Output to {output_h5.absolute()}")
        else:
            output_h5.parent.mkdir(parents=True, exist_ok=True)
        with h5py.File(str(output_h5.absolute()), "w") as out_file:
            for seq_idx in output:
                seq_grp = out_file.create_group(seq_idx)
                seq_grp.attrs["residue-order"] = MUTANT_ORDER

                # Get names for models and generate matrices
                results: Dict[str, np.ndarray] = dict.fromkeys(
                    self.models,
                    np.ones((seq_2_len[seq_idx], len(MUTANT_ORDER))) * -1,
                )

                # Write scores to matrix
                for info, pred in output[seq_idx]:
                    fromAA, postion, toAA = utils.parse_sav_str(info)
                    for name in self.models:
                        results[name][postion, REVERSE_MUTANT_ORDER[toAA]] = pred[name]

                # Transfer matrix to file
                for name in self.models:
                    seq_grp.create_dataset(name, data=results[name].T)

    @staticmethod
    def parse_logodds_input(input_file: Path):
        if not input_file.exists() or not input_file.is_file():
            raise ValueError("Logoodds input does not exist.")

        logodds_scores = dict()

        for id, emb in h5py.File(input_file, "r").items():
            id = utils.unify_seq_id(id)
            logodds_scores[id] = np.array(emb)

        return logodds_scores

    @staticmethod
    def parse_cons_input(input_file: Path):
        if not input_file.exists() or not input_file.is_file():
            raise ValueError("Conservation input does not exist.")

        conservation_scores = dict()

        for id, emb in h5py.File(input_file, "r").items():
            id = utils.unify_seq_id(id)
            conservation_scores[id] = np.array(emb)

        return conservation_scores

    @staticmethod
    def load_model(model_name: str):
        if model_name not in MODEL_PATH_DICT:
            raise ValueError(f"Requested model {model_name} not found")

        with open(MODEL_PATH_DICT[model_name], "rb") as file:
            model = pkl.load(file)
        return model

    def __generate_blosum_vec(self, mutation_gen: utils.MutationGenerator):
        vector = np.zeros((mutation_gen.total_mut_num, 1))
        idx = 0
        for (id, pos0idx, fromAA, toAA) in tqdm(
            mutation_gen.generate_all_mutations(),
            total=mutation_gen.total_mut_num,
            desc="Blosum Lookup",
            disable=not VERBOSE,
        ):
            fromAA = (
                fromAA.replace("U", "*")
                .replace("Z", "*")
                .replace("O", "*")
                .replace("X", "*")
            )
            toAA = (
                toAA.replace("U", "*")
                .replace("Z", "*")
                .replace("O", "*")
                .replace("X", "*")
            )
            vector[idx] = self.subst_matrix.get(fromAA).get(toAA)
            idx += 1
        return vector

    def __generate_prob_vec(
        self, logodds_scores, mutation_gen: utils.MutationGenerator
    ):
        vector = np.zeros((mutation_gen.total_mut_num, 1))
        idx = 0
        for (id, pos0idx, fromAA, toAA) in tqdm(
            mutation_gen.generate_all_mutations(),
            total=mutation_gen.total_mut_num,
            desc="Logodds Lookup",
            disable=not VERBOSE,
        ):
            probability = logodds_scores[id][pos0idx, mutation_gen.aa2idx[toAA]]
            vector[idx] = probability
            idx += 1
        return vector

    def __generate_cons_matrix(
        self, conservation_scores, mutation_gen: utils.MutationGenerator
    ):
        vector = np.zeros((mutation_gen.total_mut_num, NUM_MUTATION_SCORES))
        idx = 0
        for (id, pos0idx, fromAA, toAA) in tqdm(
            mutation_gen.generate_all_mutations(),
            total=mutation_gen.total_mut_num,
            desc="Conservation Lookup",
            disable=not VERBOSE,
        ):
            assert id in conservation_scores, f"{id} not in conservation scores"
            assert (
                conservation_scores[id].T.shape[0] > pos0idx
            ), f"Invalid position {pos0idx} for sequence {id}"
            vector[idx] = conservation_scores[id].T[pos0idx]
            idx += 1
        return vector

    def __generate_mutant_info(self, mutation_gen: utils.MutationGenerator):
        info = []
        for (id, pos0idx, fromAA, toAA) in tqdm(
            mutation_gen.generate_all_mutations(),
            total=mutation_gen.total_mut_num,
            desc="Info Generation",
            disable=not VERBOSE,
        ):
            info.append((id, utils.generate_sav_str(fromAA, pos0idx, toAA)))
        return info

    def __generate_input(
        self,
        mutation_gen: utils.MutationGenerator,
        conservation_dict,
        logodds_dict=None,
    ):
        input = dict()
        if self.vespa_pred:
            input["logodds_vec"] = self.__generate_prob_vec(logodds_dict, mutation_gen)
        input["blosum_vec"] = self.__generate_blosum_vec(mutation_gen)
        input["conservation_matrix"] = self.__generate_cons_matrix(
            conservation_dict, mutation_gen
        )
        return self.__generate_input_dict(**input)

    def __generate_input_dict(self, blosum_vec, conservation_matrix, logodds_vec=None):
        input = {}
        if self.vespal_pred:
            input[VESPAL] = np.concatenate((conservation_matrix, blosum_vec), axis=1)
        if self.vespa_pred:
            input[VESPA] = np.concatenate(
                (conservation_matrix, blosum_vec, logodds_vec), axis=1
            )
        return input

    def __load_models(self):
        models = {}
        if self.vespal_pred:
            models[VESPAL] = self.load_model(VESPAL)
        if self.vespa_pred:
            models[VESPA] = self.load_model(VESPA)
        return models

    def generate_predictions(
        self, mutation_gen: utils.MutationGenerator, conservations, logodds=None
    ):
        if self.vespa_pred and logodds is None:
            raise ValueError("Logodds must be present for VESPA predictions")

        input = self.__generate_input(mutation_gen, conservations, logodds)
        # Generate predictions for VESPA or for VESPA and VESPAl
        predictor_out = dict()
        if VERBOSE:
            tqdm.write("Generate Model Predictions")
        for model_name in self.models:
            # Iterate over sub-models (10 per model)
            model_dict = self.models[model_name]
            mod_input = input[model_name]

            # Iterate over data
            model_out = list()
            for fold in model_dict:
                model_out.append(model_dict[fold].predict_proba(mod_input)[:, 1])
            predictor_out[model_name] = np.mean(model_out, axis=0)
        if VERBOSE:
            tqdm.write("Predictions Done; Generate output")
        mutant_info = self.__generate_mutant_info(mutation_gen)
        # Generate output dict
        output = defaultdict(list)
        for idx, (id, mutant_str) in enumerate(mutant_info):
            output[id].append(
                (mutant_str, {name: predictor_out[name][idx] for name in self.models})
            )

        return dict(output)



from transformers import T5ForConditionalGeneration, T5EncoderModel, T5Tokenizer
from transformers import logging
logging.set_verbosity_error()
from vespa.predict.config import (
    TRANSFORMER_LINK,
    DEVICE,
    EMBED, EMBEDDING_HALF_PREC, LOGODDS
)


class ProtT5:
    def __init__(self, cache_dir):
        self.cache_dir = cache_dir

    def get_model(self, model_usage: EMBED | LOGODDS):
        if model_usage == EMBED:
            model = T5EncoderModel.from_pretrained(
                TRANSFORMER_LINK, cache_dir=self.cache_dir
            )
            if EMBEDDING_HALF_PREC:
                model = model.half()
        elif model_usage == LOGODDS:
            model = T5ForConditionalGeneration.from_pretrained(
                TRANSFORMER_LINK, cache_dir=self.cache_dir
            )
        else:
            raise NotImplementedError(
                "The intended use of ProtT5 is not implemented."
            )
        model = model.eval()
        model = model.to(DEVICE)
        return model, self.get_tokenizer()

    def get_tokenizer(self):
        tokenizer = T5Tokenizer.from_pretrained(
            TRANSFORMER_LINK, do_lower_case=False, cache_dir=self.cache_dir
        )
        return tokenizer


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['vespa', 'vespa.predict', 'vespa.scripts']

package_data = \
{'': ['*']}

install_requires = \
['biopython>=1.79,<2.0',
 'h5py>=3.4.0,<4.0.0',
 'numpy>=1.21.2,<2.0.0',
 'pyfaidx>=0.6.2,<0.7.0',
 'sentencepiece>=0.1.96,<0.2.0',
 'sklearn>=0.0,<0.1',
 'torch>=1.9.0,<2.0.0',
 'tqdm>=4.62.2,<5.0.0',
 'transformers>=4.10.2,<5.0.0']

entry_points = \
{'console_scripts': ['vespa = vespa.scripts.meta:main',
                     'vespa_conspred = vespa.scripts.conspred:main',
                     'vespa_emb = vespa.scripts.embedding:main',
                     'vespa_logodds = vespa.scripts.logodds:main',
                     'vespa_run = vespa.scripts.vespa_run:main']}

setup_kwargs = {
    'name': 'vespa-effect',
    'version': '1.0.0',
    'description': 'Predictor for single-amino acid variant effects in protein sequences (https://doi.org/10.1007/s00439-021-02411-y)',
    'long_description': '<details>\n  <summary markdown="span"> <u><b>Table of Contents</u></b> </summary>\n\n[[_TOC_]]\n\n</details>\n\n# VESPA -- **V**ariant **E**ffect **S**core **P**rediction without **A**lignments\n\n**VESPA** is a simple, yet powerful Single Amino Acid Variant (SAV) effect predictor based on embeddings of the Protein Language Model [ProtT5](https://github.com/agemagician/ProtTrans).\n\nThe single-sequence-based SAV effect prediction is set up in a multistage pipeline that includes (1) generating `ProtT5` embeddings, (2) extracting per-residue conservation predictions, (3) (optionally) extracting per-variant log-odds ratios, and (4) predicting the effect of all possible amino acid substitutions. Step (4) can be completed by either using **VESPA** with (2) and (3) as input, or by using the computationally more efficient method VESPA-light (**VESPAl**) with only step (2) as input for a small drop in prediction performance.\n\nThe specifics of **VESPA** and **VESPAl** can be found in our paper, [_Embeddings from protein language models predict conservation and variant effects_ (Marquet et al., 2021)](https://doi.org/10.1007/s00439-021-02411-y). The performance of **VESPA** when evaluated against SOTA methods can be seen below.\n<div align="center"><img src="images/point_main_0823.png" width="75%" height="75%"/></div>\n\n## Precomputed VESPA and VESPAl predictions\n\nPrecomputed **VESPA** and **VESPAl** predictions are currently available for 39 _DMS experiments_ here:  [_Supplementary file 3 of (Marquet et al., 2021)_](https://doi.org/10.1007/s00439-021-02411-y). Furthermore, **VESPAl** predictions are available for the [_human proteome_](https://doi.org/10.5281/zenodo.5905863) (downloaded 22/01/17), and for the [__fly (drosophila melanogaster) proteome__](https://doi.org/10.5281/zenodo.6322184) (downloaded 22/03/01).\n\n## Usage\n\nThe preferred method to install **VESPA** is via pip:\n\n```bash\npip install vespa-effect\n```\n\n-----\n\n### Input Files\n\n**Required**: A single `FASTA` file containing all your wildtype sequences (note: this file can contain any number of sequences).\n\n**Optional**: If you are only interested in a subset of possible mutations (specific mutations), you can add `-m mutations.txt` to the code line in [Quickstart](#Quickstart) (note: per default all mutations are considered). Click [`here`](#mutations-file) to head to the file format explanation of `mutations.txt`.\n\n-----\n\n### Quickstart\n\nFor simplicity of this guide, we will assume a folder containing all data: f.e., the `FASTA` file is placed at `data/sequences.fasta` and the (optional) `mutations.txt` at `data/mutations.txt`.\n\nAfter installing the repository, you can run the following:\n\n```bash\nvespa data/sequences.fasta --prott5_weights_cache data/cache\n```\n\nBy default, this runs **VESPAl** for all possible mutations. It will generate a new folder _vespa_run_directory_ in your current working directory. Within this folder you will find two files containing input features (embeddings, conservation prediction) and an _output_ folder with .csv files containing **VESPAl** predictions for each sequence with all possible mutations. More details can be found under [output](#vespa-and-vespal-output).\n\n**WARNING** Creating embeddings requires a powerful GPU (we recommend at least 12GB of VRAM). The same applies to running **VESPA**. For more details, see [Step 1 on generating embeddings](#step-1-extracting-prott5-embeddings) and [Step 3 on extracting log-odds ratios](#step-3-log-odds-ratio-of-masked-marginal-probabilities).\n\n**Optional**:\n\n- If you already have precomputed embeddings available, you can use those instead of generating new ones by adding `--use_existing_embeddings` followed by the file location. Then running **VESPAl** will not require a GPU.\n- If you have a GPU available you can generate **VESPA** predictions by adding `--vespa` to the code line above. After running **VESPA** you will find another file in your folder _vespa_run_directory_ containing log-odds ratios as additional input.\n- `--prott5_weights_cache` followed by a folder path specifies the location of a caching folder (or preloaded ProtT5 weigths). Instead of downloading the weights for every run of **VESPA** and/or **VESPAl**, the weights will be downloaded into this folder or just reused if already present.\n- In case you want to output a single [.h5 file](#.h5 files) in addition to the default .csv files with predictions, add `--h5_output` followed by the output file location. If you are only interested in an .h5 file, add `--no_csv`.\n\nBelow you can find information on how to run each step of **VESPA** and **VESPAl** individually.\nRunning the vespa script will automatically the substeps below for you (with the optional specifications you of your choice). Use the substeps if you are interested in a particular intermediate result.\n\n-----\n\n### The substeps of VESPA and VESPAl\n\nThe following steps can be run individually and for any number of sequences contained in a `FASTA` file.\n\n#### Step 1: Extracting ProtT5 embeddings\n\nTo run **VESPA** and/or **VESPAl** to obtain SAV predictions, you will need the `ProtT5` embeddings of your sequences. If you have a GPU, you can use the included embedding script on your own machine to generate your protein embeddings:\n\n```bash\nvespa_emb data/sequences.fasta -o data/embeddings.h5  --prott5_weights_cache data/cache\n\n```\n\nAs you can see, you need to specify an output file location (`-o`), and the `FASTA` file location of your input sequences.\n\n**Optional**: `--prott5_weights_cache` specifies a caching directory to store `ProtT5` weights. If not present, the line above will download the weights to your current working directory for every run of **VESPA** and/or **VESPAl**.\n\n#### Step 2: Conservation Prediction\n\n**VESPA** and **VESPAl** take per-residue conservation predictions as input. To generate them, run the following (in the `VESPA` folder):\n\n```bash\nvespa_conspred data/embeddings.h5 -o data/conspred.h5\n```\n\nThe input for our conservation predictor are the `ProtT5` embeddings. In your output file (`-o`), you will find 9-state conservation probabilities (per-residue) needed as input for the models. For more details on the conservation prediction please see the paper mentioned above.\n\n**Optional**: In case you are interested in generating a file that contains the predicted conservation classes instead of the assigned class probabilities, add `--output_classes` above.\n\n#### Step 3: Log-odds ratio of masked marginal probabilities\n\nThis step requires a GPU! The two models **VESPA** and **VESPAl** differ in their predictive performance but also in the required input. If you are only interested in running **VESPAl**, you can skip this step.\n\nTo generate the log-odds ratios of masked marginal probabilities required for **VESPA**, run (in the `VESPA` folder):\n\n```bash\nvespa_logodds data/sequences.fasta -o data/logodds.h5 \n```\n\n**Optional**:\n\n- To run the log-odds script including the mutations file, add `-m data/mutations.txt` to the code line above.\n- Per default the command above generates the [`.h5` file](#h5-files) required for the subsequent steps. We provide two options to output a human readable version of the log-odds scores by adding `--single_csv data/single_logodds.csv` or `--csv_dir data/csv_dir/` at the end of the code line above. The first option generates a single csv file with all sequences and SAVs. For large sets, we recommend to use the second option, which outputs multiple csv files separated by sequence ID of the given `FASTA` file. The format of the .csv files is described [below](#log-odds-ratio-output).\n\n#### Step 4: Run **VESPA** and/or **VESPAl**\n\nNow you have all the data required to run **VESPA** and/or **VESPAl**. Per default, the `vespa` script will predict SAV effects for **VESPAl** only. To explicitly enable **VESPA** add `--vespa` (which will run both models), and to disable **VESPAl** add `--no-vespal`. To generate predictions, execute the following code (in the `VESPA` folder):\n\n- **Both** (if you computed the conservation prediction and the log-odds):\n\n    ```bash\n    vespa_run --vespa data/conspred_probs.h5 data/sequences.fasta --T5_input data/logodds.h5 --output predictions/\n    ```\n\n- **Only VESPA** (if you computed the conservation prediction and the log-odds):\n\n    ```bash\n    vespa_run --vespa --no-vespal data/conspred_probs.h5 data/sequences.fasta -T5_input data/logodds.h5 --output predictions/\n    ```\n\n- **Only VESPAl** (if you only computed the conservation prediction):\n\n    ```bash\n    vespa_run data/conspred_probs.h5 data/sequences.fasta --output predictions/\n    ```\n\n**Optional**: In case you want to output a single [.h5 file](#.h5 files) in addition to the default .csv files with predictions, add `--h5_output` followed by the output location. If you are only interested in an .h5 file, add `--no_csv` and drop `--output predictions/`. More details can be found under [output](#vespa-and-vespal-output).\n\n**Note:** The format of the results file is described below at [VESPA and VESPAl output](#vespa-and-vespal-output).\n\n-----\n\n### Additional Information\n\n#### Extracting raw reconstruction probabilities\n\nYou might be interested in extracting the raw reconstruction probabilities for each mutation position from `ProtT5`. These raw reconstruction probabilities are used to calculate the log-odds ratio and is explained in more detail in the corresponding publication.\nTo do so, use:\n\n```bash\nvespa_logodds -r --reconstruction_output data/reconstruction_probas data/sequences.fasta -o data/logodds.h5 \n```\n\n**Optional**: Add a mutations file.\n\nThe generated datasets in the [`.h5` file](#h5-files) contain probability vectors that determine the reconstruction probabilities for all amino acids sorted according to the `MUTATION_ORDER` in `config.py`. If a particular mutation was not computed (i.e., the position was not present in the `mutation.txt`) it contains -1.\n\n#### File Specifications\n\nThis section describes relevant file formats for **VESPA** and **VESPAl**:\n\n##### Mutations file\n\nA simple text file with a protein ID and one mutation per-line (i.e., mutations separated by \\n for newline).\nEvery mutation should be specified by `<PROTEIN_ID>_<SAV-String>` separated by an underscore. The default `<SAV-String>` has the following format `<fromAA><0-basedSequencePosition><toAA>`.\n\nThe sequence ID needs to be equivalent to the one in the sequence `FASTA` file.\nThe SAV string has the format: `<Original Amino Acid><Position><Replacement Amino Acid>`\n\nExample:\n\n```txt\nENSP00000355206_I0L\nENSP00000355206_I0V\nENSP00000355206_I0L\nENSP00000355206_I0K\nENSP00000355206_I0T\n```\n\n**Optional:** In case you have a file with 1-based mutations, add the flag `--one_based_mutations`.\n\n##### `.h5` files\n\nMultiple scripts generate `.h5` files. These files follow the [`hdf5`-standard](https://www.hdfgroup.org/solutions/hdf5)\nand can be processed in python using the library [`h5py`](https://www.h5py.org/). Generally, the files are segmented into datasets that can be accessed using the protein accession in the `FASTA` file. The datasets are explained in more detail in the respective _output sections_.\n\n##### Conservation output\n\nThe conservation prediction will output an .h5 file with the predicted 9-state conservation probabilities. For each ID of the `FASTA` file, the respective dataset contains a matrix of size 9xL, with L being the length of the protein sequence, and 9 being the predicted conservation class (index 0 = very variable; index 8 = very conserved).\n\nAdditionally, you can extract a file with per-residue conservation class predictions for each sequence directly.\n\n##### Log-odds ratio output\n\nThe default output is an .h5 file that contains one Lx20 matrix per sequence. -1 means no log-odds ratio was computed for the respective mutation.\n\nWhen running `--single_csv data/single_logodds.csv`, the log-odds ratio output file contains all mutations (or the mutations determined by the mutations file) as `<PROTEIN_ID>_<SAV-String>`, followed by a `;` and the log-odds ratio for a particular mutation per line.\n\nExample:\n\n```csv\nB3VI55_LIPSTSTABLE_A438Q;0.053829677402973175\nB3VI55_LIPSTSTABLE_A438N;0.061238136142492294\nB3VI55_LIPSTSTABLE_A438Y;0.012603843584656715\nB3VI55_LIPSTSTABLE_A438M;0.012212143279612064\nB3VI55_LIPSTSTABLE_A438H;0.0241163931787014\nB3VI55_LIPSTSTABLE_A438W;0.004520446062088013\n```\n\nIn case the predictions are written into a directory, e.g. by specifying `--csv_dir` in [Step 3](#step-3-log-odds-ratio-of-masked-marginal-probabilities), the script will create one file per sequence, named by sequence ID (note: the ID will be normalized, i.e, each special character will be replaced by `_`).\nThe file will contain one `<Mutation-String>;score` per-line.\n\nExample file called `B3VI55_LIPSTSTABLE`:\n\n```csv\nA438Q;0.053829677402973175\nA438N;0.061238136142492294\nA438Y;0.012603843584656715\nA438M;0.012212143279612064\nA438H;0.0241163931787014\nA438W;0.004520446062088013\n```\n\n##### VESPA and VESPAl output\n\nThe default output of **VESPA** and/or **VESPAl** will generate one .csv file per protein in the output folder of `vespa_run_directory`. If you specified an output directory with `--output`, you will find the .csv files there.\nTo circumvent naming issues due to long sequence ID\'s, the csv files will be numbered by sequence occurrence in the `FASTA` file. A lookup file `map.json` will be created in the output directory containing a dictionary mapping from number to sequence ID.\n\nExample `map.json`:\n\n```json\n{    "0": "B3VI55_LIPSTSTABLE",\n    "1": "BF520_ENV",\n    "2": "BG_STRSQ",\n    "3": "BG505_ENV",\n    "4": "HG_FLU",\n    "5": "MTH3_HAEAESTABILIZED"}\n```\n\nThe individual files will contain rows with the mutations (format: `<fromAA><0-basedSequencePosition><toAA>`) along with the respective predictions of **VESPA** and/or **VESPAl**.\n\nExample `0.csv`:\n\n```csv\nMutant;VESPAl;VESPA\nM0A;0.4457732174287125;0.3520255108578212\nM0L;0.3191178420567241;0.2717188481387661\nM0G;0.5355136080284415;0.4110670843315182\nM0V;0.3594337197937546;0.2971971641898669\nM0S;0.4457732174287125;0.35202555423053\nM0R;0.4457732174287125;0.35202621644931126\n```\n\nIf you chose to generate an .h5 output, the ID\'s will be the original `FASTA` ID\'s. The file contains one 20xL-shaped matrix per sequence-ID. L is the length of the respective protein length, and 20 is the possible number of amino acid variants (including self). The order of amino acids is determined by `MUTANT_ORDER` in `predict/config.py` (`ALGVSREDTIPKFQNYMHWC`). Empty fields that were not calculated/specified (f.e., wildtype substitutions) contain a -1.\n\n-----\n\n## Development Roadmap\n\n- [ ] Write comprehensive tests\n- [x] Publish pypi package\n- [x] Install from github release\n- [ ] Contributing guidelines\n\n-----\n\n## Installation from current Github Release\n\n**WARNING Experimental**: To install the current release from github you can use:\n\n```bash\npython -m pip install https://github.com/Rostlab/VESPA/releases/download/v0.9.0-beta/vespa-0.9.0b0.tar.gz\n```\n\n-----\n\n## Cite\n\nIf you want to credit us, feel free to cite\n\nMarquet, C., Heinzinger, M., Olenyi, T. et al. Embeddings from protein language models predict conservation and variant effects. Hum Genet (2021). <https://doi.org/10.1007/s00439-021-02411-y>\n\n```Bibtex\n@article{Marquet2021,\n  doi = {10.1007/s00439-021-02411-y},\n  url = {https://doi.org/10.1007/s00439-021-02411-y},\n  year = {2021},\n  month = dec,\n  publisher = {Springer Science and Business Media {LLC}},\n  author = {C{\\\'{e}}line Marquet and Michael Heinzinger and Tobias Olenyi and Christian Dallago and Kyra Erckert and Michael Bernhofer and Dmitrii Nechaev and Burkhard Rost},\n  title = {Embeddings from protein language models predict conservation and variant effects},\n  journal = {Human Genetics}\n}\n',
    'author': 'Tobias O',
    'author_email': 'tobias.olenyi@tum.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Rostlab/VESPA',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<3.11',
}


setup(**setup_kwargs)



#!/usr/bin/env python
"""
 Copyright (C) 2021 Rostlab
 
 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU Affero General Public License as
 published by the Free Software Foundation, either version 3 of the
 License, or (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Affero General Public License for more details.
 
 You should have received a copy of the GNU Affero General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

# -*- coding: utf-8 -*-

from collections import defaultdict
from pathlib import Path
from typing import DefaultDict, Dict, Iterable, Iterator, Tuple
from pyfaidx import Fasta
import numpy as np
import re

NON_WORD_RE = re.compile("[^a-zA-Z0-9]")

from vespa.predict.config import *


def unify_seq_id(id: str) -> str:
    return NON_WORD_RE.sub("_", id)


def parse_fasta_input(input_file: Path, split_char="!", id_field=0):
    def key_function(header):
        if "|" in header:
            header = header.split("|")[1]
        else:
            header = header.split(split_char)[id_field]
        # replace tokens that are mis-interpreted when loading h5
        header = unify_seq_id(header)
        return header

    sequences = dict()

    seqs_file = Fasta(
        str(input_file.absolute()),
        key_function=key_function,
        sequence_always_upper=True,
    )
    for rec in seqs_file:
        seq = str(rec[:])
        sequences[rec.name] = str(
            seq
            if not REPLACE_RARE_AA
            else seq.replace("U", "X").replace("Z", "X").replace("O", "X")
        )
    return sequences


def generate_sav_str(fromAA, pos0idx, toAA):
    return f"{fromAA}{pos0idx}{toAA}"


def generate_mutation_str(id, fromAA, pos0idx, toAA):
    return f"{id}:{generate_sav_str(fromAA, pos0idx, toAA)}"


def parse_mutation_str(mutation: str) -> Tuple[str, str, int, str]:
    parts = NON_WORD_RE.split(mutation.strip())
    sav = parse_sav_str(parts[-1])
    return "_".join(parts[:-1]), *sav


def parse_sav_str(sav: str) -> Tuple[str, int, str]:
    fromAA = sav[0]
    toAA = sav[-1]
    pos0idx = int(sav[1:-1])
    return fromAA, pos0idx, toAA


def pred_to_str(pred: np.float64):
    if RESULT_PREC > 0:
        fmt = f".{RESULT_PREC}f"
        res = f"{pred:fmt}"
    else:
        res = f"{pred}"
    return res


class MutationGenerator:
    def __init__(
        self,
        sequence_dict: Dict[str, str],
        file_path: Path = None,
        order=MUTANT_ORDER,  # ignore X for list of viable mutations
        one_based_file=False,
        rare_aas=False,
    ) -> None:
        self.one_based = one_based_file
        self.sequence_dict = sequence_dict
        self.load_from_file = False
        self.order = order
        if file_path is not None:
            self.load_from_file = True
            self.mutation_file_path = file_path
            self.mutation_dict = self._parse_mutations_from_file(
                self.mutation_file_path, self.one_based
            )
        self.total_mut_num = self.__calc_total_mutation_num()
        self.aa2idx = self._get_aa2idx()

    def _get_aa2idx(self):
        return {aa: idx for idx, aa in enumerate(self.order)}

    def __calc_total_mutation_num(self):
        num = 0
        if self.load_from_file:
            for id in self.mutation_dict:
                num += len(self.mutation_dict[id])
        else:
            for id in self.sequence_dict:
                num += len(self.sequence_dict[id]) * (len(self.order) - 1)
                num += self.sequence_dict[id].count("X")

        return num

    def total_mutation_num(self):
        return self.total_mut_num

    def mutation_num_id(self, id):
        """
        Get the number of requested mutations per sequence id

        Args:
            id (str): Fasta Sequence-Id

        Returns:
            int : number of requested mutations per sequence
        """
        if self.load_from_file:
            seq = self.mutation_dict.get(id, None)
            return len(seq) if seq is not None else 0
        else:
            seq = self.sequence_dict.get(id, None)
            return len(seq) * (len(self.order) - 1) if seq is not None else 0

    @staticmethod
    def _parse_mutations_from_file(mutation_file: Path, one_based_file: bool):
        """
        Read the mutations file into a dictionary

        Args:
            mutation_file (Path): Path to the mutations.txt
            one_based_file (bool): Determines if coordinates in the file start with 1

        Returns:
            Dict[str, List[Tuple[int, str,str ]]]: Dictionary mapping sequence id to a  sequence of mutation positions followed by fromAA and toAA
        """

        mutations_dict = defaultdict(list)
        with mutation_file.open("r") as mutations:
            for line in mutations:
                id, fromAA, pos0idx, toAA = parse_mutation_str(line)
                pos0idx = pos0idx if not one_based_file else (pos0idx - 1)
                mutations_dict[id].append((pos0idx, fromAA, toAA))
        return dict(mutations_dict)

    def __generate_all_mutations_from_file(self) -> Iterator[Tuple[str, ...]]:
        """
        Generator function;
        Yield all mutations for the sequences as listed in the given mutations file.

        Yields:
            Iterator[Tuple[str, ...]]: Generator that yields sequence id, followed by mutation information (see `generate_all_seq_mutations_from_file`).
        """
        for id in self.sequence_dict:
            if id in self.mutation_dict:
                for (
                    pos0idx,
                    fromAA,
                    toAA,
                ) in self.__generate_all_seq_mutations_from_file(id):
                    yield id, pos0idx, fromAA, toAA

    def __generate_all_seq_mutations_from_file(self, id: str):
        for mutation in self.mutation_dict[id]:
            yield mutation

    def __generate_all_mutations(self) -> Iterator[Tuple[str, ...]]:
        """Generate all viable mutations for every sequence in the given fasta file.

        Returns:
            Iterable[Tuple[str, ...]]: sequence-id followed by mutations (see `generate_all_seq_mutations`)

        Yields:
            Iterator[Iterable[Tuple[str, ...]]]: [description]
        """
        for id, sequence in self.sequence_dict.items():
            for pos0idx, fromAA, toAA in self.__generate_all_seq_mutations(id):
                yield id, pos0idx, fromAA, toAA

    def __generate_all_seq_mutations(self, id: str) -> Iterator[Tuple[int, str, str]]:
        """
        Generator function;
        Generate all possible mutations for a given fasta sequence-id.

        Args:
            id (str): the sequence id in the given fasta file.


        Yields:
            Iterator[Tuple[int, str, str]]: Generator for all possible mutations at each sequence postition; format: 0-based index, fromAA, toAA. Does not gernate self mutations.
        """

        sequence = self.sequence_dict[id]
        for pos0idx in range(len(sequence)):
            for toAA in self.order:
                fromAA = (
                    sequence[pos0idx]
                    if not REPLACE_RARE_AA
                    else sequence[pos0idx]
                    .replace("U", "X")
                    .replace("Z", "X")
                    .replace("O", "X")
                )
                if fromAA != toAA:
                    yield pos0idx, fromAA, toAA

    def generate_all_mutations(self):
        """Generate all mutations either from a mutations file if given or all mutations.

        Returns:
            Iterable[Tuple[str, int, str, str]]: Sequence-Id, 0-based sequence postion, wild-type AA, mutant AA
        """
        if self.load_from_file:
            return self.__generate_all_mutations_from_file()
        else:
            return self.__generate_all_mutations()

    def generate_mutations_for_seq(self, seq_id):
        """
        Retrun generator for all mutations for the given sequence id either from file if given or just all possible mutations

        Args:
            seq_id (str): The Fasta id of the required sequence

        Returns:
            Generator[Tuple[int, str, str]] : Generator that generates the relevant sequence 0-based position, fromAA, toAA.
        """
        if self.load_from_file:
            return self.__generate_all_seq_mutations_from_file(seq_id)
        else:
            return self.__generate_all_seq_mutations(seq_id)

    def generate_mutation_mask(self, include_x_pos=False):
        """
        Generator function.
        Generates a mutation mask; essentially a numpy array per amino acid position with 1 on positions where scores should be used. This is mostly relevant if not all mutations per position should be considered.

        Yields:
            Dict[np.array]: dictionary that maps the specific sequence position to the mutations that should be considered.
        """
        old_id = None
        aa2idx = self.aa2idx
        if include_x_pos:
            aa2idx["X"] = len(self.aa2idx)
        idx_filter = defaultdict(lambda: np.zeros(len(aa2idx)))
        # Note: relies on the generators producing mutations in order by id
        # Essentially iterate over mutations until you see new id then return the combined mask.
        for id, pos0idx, fromAA, toAA in self.generate_all_mutations():
            if old_id is None:
                old_id = id
            if id != old_id:
                yield old_id, idx_filter
                old_id = id
                idx_filter = defaultdict(lambda: np.zeros(len(aa2idx)))

            idx_filter[pos0idx][self.aa2idx[toAA]] = 1
        yield old_id, idx_filter

    def generate_mutation_file(self, output_file: Path, percentage=0.2):
        # Sample percentage indices
        indices = np.random.choice(
            self.total_mut_num, (int)(self.total_mut_num * percentage)
        )

        with output_file.open("w") as mutations:
            for idx, (id, pos0idx, fromAA, toAA) in enumerate(
                self.__generate_all_mutations()
            ):
                if idx in indices:
                    mutations.write(
                        f"{generate_mutation_str(id, fromAA, pos0idx, toAA)}\n"
                    )
